from ROOT import RDataFrame
import numpy as np

from . import logger

#TODO: add logger
_logger = logger.Logger(__name__, 'info').get_logger()

def get_tree_branch_as_np(tree_name: str, 
                          root_file: str, 
                          branch_names:str, 
                          filter_string: str = None) -> dict[str, np.ndarray]:
    '''
    Loads a TTree from a .root File, applies a filter and returns defined 
    branches (columns) as a dictionary.
    The column names are the keys of the dictionary. 
    If the column contains only integers, a 1D Numpy array
    is the value. If the column contains C++ Vectors, it is cast to a python 
    list. The value is a 1D
    Numpy array of objects (python lists).
    It is recommended to filter the data needed, since loading to memory it 
    quite slow. A "filter_string" can
    be provided. The content must be C++ code. The name of the variables must 
    be the name of the branches.
    eg: filter_string = 'NSignals >= 2 && FrameIndex == 8'
    Args:
        tree_name: name of a TTree
        root_file: path of the .root file
        branch_names: list of strings containing names of branches to load
        filter_string: string for a filter 
    Returns:
        Dictionary of numpy arrays
    '''
    #TODO: add better exception handling
    try:
        df = RDataFrame(tree_name,root_file)
        print(f'Tree loaded succesfully: \n {df.Describe()}')
    except:
        print('Tree could not be loaded.')
        return 0
    #get a dictionary of types
    column_names = list(df.GetColumnNames())
    column_type = {col: df.GetColumnType(col) for col in column_names}
    
    #Apply a filter string. It is recommended to use one to avoid loading unneccesary data.
    #The content must be C++ code. The name of the variables must be the name of the branches.
    if filter_string is not None:
        try:
            df = df.Filter(filter_string)
        except:
            print('Filter {filter_string} could not be applied')
            return 0
    
    #now get the branches in a numy array and transform C++ Objects to Python Lists
    output = df.AsNumpy(columns=branch_names)
    for key in output.keys():
        branch_entry_type = column_type[key]
        if 'RVec' in branch_entry_type:
            for index, vec in enumerate(output[key]):
                content = list(vec)
                output[key][index] = content
    return output

def get_count_map_from_pixel_index(inp_dict: 
                                   dict[str, np.ndarray]) -> np.ndarray:
    '''
    Takes a dictionary (the output of get_tree_branch_as_np()) with 'Column' 
    and 'Row' keys and returns a 64 by 64 map with the counts.
    Args:
        inp_dict: dictionary with a list of lists for 'Column' and 'Row'
    Returns:
        64 by 64 numpy array
    '''
    #TODO: add better exception handling
    if type(inp_dict) != dict:
        print('Not a dictionary!')
        return
    output = np.zeros((64,64))
    keys = inp_dict.keys()
    if 'Column' not in keys or 'Row' not in keys :
        print(f'No Pixel indizes in keys of dictionary {inp_dict.keys()}')
        return
    cols = inp_dict['Column']
    rows = inp_dict['Row']
    if len(cols) != len(rows):
        print('Something is wrong with the event tree.')
        return
    else:
        #loop through events
        for event_no in range(len(cols)):
            #get the indizes for that event
            cols_event = cols[event_no]
            rows_event = rows[event_no]
            if len(cols_event) != len(rows_event):
                print('Something is wrong with the event tree.')
                return
            else:
                #loop throug indices for that event and fill map
                for i in range(len(cols_event)):
                    col = cols_event[i]
                    row = rows_event[i]
                    output[col,row] += 1
    return output


import gc
import os

import numpy as np

from . import parallel_funcs
from . import logger

_logger = logger.Logger(__name__, 'info').get_logger()

def get_avg_over_frames(data: np.ndarray) -> np.ndarray:
    '''
    Calculates the average over the frames in data.
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
    
    Returns:
        np.array (column_size, nreps, row_size)
    '''
    if np.ndim(data) != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    return parallel_funcs.nanmean(data, axis=0)

def get_avg_over_nreps(data: np.ndarray) -> np.ndarray:
    '''
    Calculates the average over the nreps in data.
    Args:
        data: np.array in shape (nframes, column_size, nreps, row_size)
    Returns:
        np.array in shape (nframes, column_size, row_size)
    '''
    if np.ndim(data) != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    return parallel_funcs.nanmean(data, axis=2)

def get_avg_over_frames_and_nreps(data : np.ndarray,
                                  avg_over_frames: np.ndarray = None,
                                  avg_over_nreps: np.ndarray = None) -> np.ndarray:
    '''
    Calculates the average over the frames and nreps in data. If avg_over_frames
    or avg_over_nreps are already calculated they can be passed as arguments to
    save computation time.
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
        avg_over_frames: (optional) np.array (column_size, nreps, row_size)
        avg_over_nreps: (optional) np.array (nframes, column_size, row_size)
    Returns:
        np.array (column_size, row_size)
    '''
    if np.ndim(data) != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    
    if avg_over_frames is None and avg_over_nreps is None:
        return parallel_funcs.nanmean(parallel_funcs.nanmean(data, axis=0), axis = 2)
    
    if avg_over_frames is not None and avg_over_nreps is not None:
        if np.ndim(avg_over_frames) != 3 or np.ndim(avg_over_nreps) != 3:
            _logger.error('Input avg_over_frames or avg_over_nreps is not a 3D array.')
            raise ValueError('Input avg_over_frames or avg_over_nreps is not a 3D array.')
        if avg_over_frames.shape[1] < avg_over_nreps.shape[0]:
            return parallel_funcs.nanmean(avg_over_frames, axis=1)
        else:
            return parallel_funcs.nanmean(avg_over_nreps, axis=0)
    else: 
        if avg_over_nreps is not None:
            if np.ndim(avg_over_nreps) != 3:
                _logger.error('Input avg_over_nreps is not a 3D array.')
                raise ValueError('Input avg_over_nreps is not a 3D array.')
            return parallel_funcs.nanmean(avg_over_nreps, axis=0)
        else:
            if np.ndim(avg_over_frames) != 3:
                _logger.error('Input avg_over_frames is not a 3D array.')
                raise ValueError('Input avg_over_frames is not a 3D array.')
            return parallel_funcs.nanmean(avg_over_frames, axis=1)

def get_rolling_average(data: np.ndarray, window_size: int) -> np.ndarray:
    '''
    Calculates a rolling average over window_size
    Args:
        data: 1D np.array
        window_size: int
    Returns:
        1D np.array
    '''
    weights = np.repeat(1.0, window_size) / window_size
    # Use 'valid' mode to ensure that output has the same length as input
    return np.convolve(data, weights, mode='valid')

def load_npy_files(folder: str) -> dict:
    '''
    Looks for .npy arrays in folder and returns them all as a
    dictionary with numpy arrays as values
    Args:
        folder: folder path
    Returns:
        dictionary of np.arrays
    '''
    # Get a list of all .npy files in the folder
    files = [f for f in os.listdir(folder) if f.endswith('.npy')]
    if len(files) == 0:
        print(f'No .npy files found in folder {folder}')
        return None
    # Load each file into a numpy array and store it in a dictionary
    arrays = {}
    for file in files:
        # Remove the .npy extension from the filename
        name = os.path.splitext(file)[0]
        # Load the file and store it in the dictionary
        arrays[name] = np.load(os.path.join(folder, file), allow_pickle=True)
    return arrays

def sort_with_indices(arr: np.ndarray) -> np.ndarray:
    '''
    Sorts array in descending order and returns the indices.
    Args:
        arr: 1D np.array
    Returns:
        1D np.array
    '''
    if np.ndim(arr) != 1:
        _logger.error('Input data is not a 1D array.')
        raise ValueError('Input data is not a 1D array.')
    indexed_arr = np.column_stack((np.arange(len(arr)), arr))
    sorted_indices = indexed_arr[np.argsort(indexed_arr[:, 1])[::-1]][:, 0]
    return sorted_indices.astype(int)

def get_array_from_file(folder: str, filename: str) -> np.ndarray:
    return np.load(os.path.join(folder, filename), allow_pickle=True)


import gc
import os
from datetime import datetime

import numpy as np

from . import logger
from . import analysis_funcs as af
from . import analysis as an
from . import params as pm
from . import fitting as fit

class OffNoi():
    _logger = logger.Logger('nproan-offnoi', 'debug').get_logger()

    def __init__(self, prm_file):
        self.load(prm_file)
        self._logger.info('OffNoi object created')

    def load(self, prm_file):
        self.params = pm.Params(prm_file)
        self.params_dict = self.params.get_dict()
        
        #common parameters
        self.results_dir = self.params_dict['common_results_dir']
        self.column_size = self.params_dict['common_column_size']
        self.row_size = self.params_dict['common_row_size']
        self.key_ints = self.params_dict['common_key_ints']
        self.bad_pixels = self.params_dict['common_bad_pixels']
        #offnoi parameters
        self.bin_file = self.params_dict['offnoi_bin_file']
        self.nreps = self.params_dict['offnoi_nreps']
        self.nframes = self.params_dict['offnoi_nframes']
        self.nreps_eval = self.params_dict['offnoi_nreps_eval']
        self.comm_mode = self.params_dict['offnoi_comm_mode']
        self.thres_mips = self.params_dict['offnoi_thres_mips']
        self.thres_bad_frames = self.params_dict['offnoi_thres_bad_frames']
        self.thres_bad_slopes = self.params_dict['offnoi_thres_bad_slopes']

        #directories, they will be created in calculate()
        timestamp = datetime.now().strftime('%Y%m%d-%H%M%S')
        filename = os.path.basename(self.bin_file)[:-4]
        self.common_dir = os.path.join(
            self.results_dir, timestamp + '_' + filename)
        self.step_dir = os.path.join(self.common_dir, 
            f'offnoi_{self.nreps}reps_{self.nframes}frames')

        self._logger.info(f'Parameters loaded:')
        self.params.print_contents()

    def calculate(self):   
        #create the directory for the data
        os.makedirs(self.common_dir, exist_ok=True)
        self._logger.info(f'Created common directory for data: {self.common_dir}')
        #now, create the working directory for the offnoi step
        os.makedirs(self.step_dir, exist_ok=True)
        self._logger.info(f'Created working directory for offnoi step: {self.step_dir}')
        # and save the parameter file there
        self.params.save(os.path.join(self.step_dir, 'parameters.json'))

        data = an.get_data(self.bin_file, self.column_size, self.row_size, self.key_ints, self.nreps, self.nframes)
        gc.collect()
        #delete nreps_eval from data
        if self.nreps_eval:
            data = an.exclude_nreps_eval(data, self.nreps_eval)
            self._logger.debug(f'Shape of data: {data.shape}')
        #set values of all frames and nreps of bad pixels to nan
        if self.bad_pixels:
            data = an.set_bad_pixellist_to_nan(data, self.bad_pixels)
        #deletes bad frames from data
        if self.thres_bad_frames != 0:
            data = an.exclude_bad_frames(data, self.thres_bad_frames, self.step_dir)
            self._logger.debug(f'Shape of data: {data.shape}')
        #deletes frames with mips above threshold from data
        if self.thres_mips != 0:
            data = an.exclude_mips_frames(data, self.thres_mips)
            self._logger.debug(f'Shape of data: {data.shape}')
        #calculate offset_raw on the raw data and save it
        avg_over_frames = af.get_avg_over_frames(data)
        np.save(os.path.join(self.step_dir, 'offset_raw.npy'),
                avg_over_frames)
        #calculate offset and save it
        avg_over_frames_and_nreps = af.get_avg_over_frames_and_nreps(data)
        np.save(os.path.join(self.step_dir, 'offset.npy'),
                avg_over_frames_and_nreps)
        #offset the data and correct for common mode if necessary
        data -= avg_over_frames[np.newaxis,:,:,:]
        del avg_over_frames
        gc.collect()
        if self.comm_mode is True:
            an.correct_common_mode(data)
        #calculate rndr signals and save it
        avg_over_nreps = af.get_avg_over_nreps(data)
        np.save(os.path.join(self.step_dir, 'rndr_signals.npy'),
                avg_over_nreps)
        #calculate fitted offset and noise and save it (including fit errors)
        fit_unbinned = fit.get_unbinned_fit_gauss(avg_over_nreps)
        fit_curve_fit = fit.get_fit_gauss(avg_over_nreps)
        np.save(os.path.join(self.step_dir, 'offnoi_fit_unbinned.npy'), fit_unbinned)
        np.save(os.path.join(self.step_dir, 'offnoi_fit.npy'), fit_curve_fit)
        if self.thres_bad_slopes != 0:
            bad_slopes_pos, bad_slopes_data, bad_slopes_value = an.get_bad_slopes(data, self.thres_bad_slopes, self.step_dir)
            np.save(os.path.join(self.step_dir, 'bad_slopes_pos.npy'), bad_slopes_pos)
            np.save(os.path.join(self.step_dir, 'bad_slopes_data.npy'), bad_slopes_data)
            np.save(os.path.join(self.step_dir, 'bad_slopes_value.npy'), bad_slopes_value)

class Filter():

    _logger = logger.Logger('nproan-filter', 'debug').get_logger()

    def __init__(self, prm_file, offnoi_dir):
        self.load(prm_file, offnoi_dir)
        self._logger.info('Filter object created')

    def load(self, prm_file, offnoi_dir):
        self.params = pm.Params(prm_file)
        self.params_dict = self.params.get_dict()
        #common parameters
        self.results_dir = self.params_dict['common_results_dir']
        self.column_size = self.params_dict['common_column_size']
        self.row_size = self.params_dict['common_row_size']
        self.key_ints = self.params_dict['common_key_ints']
        self.bad_pixels = self.params_dict['common_bad_pixels']
        #filter parameters
        self.bin_file = self.params_dict['filter_bin_file']
        self.nreps = self.params_dict['filter_nreps']
        self.nframes = self.params_dict['filter_nframes']
        self.nreps_eval = self.params_dict['filter_nreps_eval']
        self.comm_mode = self.params_dict['filter_comm_mode']
        self.thres_mips = self.params_dict['filter_thres_mips']
        self.thres_event = self.params_dict['filter_thres_event']
        self.use_fitted_offset = self.params_dict['filter_use_fitted_offset']
        self.thres_bad_frames = self.params_dict['filter_thres_bad_frames']
        self.thres_bad_slopes = self.params_dict['filter_thres_bad_slopes']

        #directories
        #set self.common_dir to the parent directory of offnoi_dir
        self.common_dir = os.path.dirname(offnoi_dir)
        self.step_dir = None
        
        self._logger.info(f'Parameters loaded:')
        self.params.print_contents()
        
        self._logger.info('Checking parameters in offnoi directory')
        #look for a json file in the offnoi directory 
        if (not self.params.same_common_params(offnoi_dir)) \
            or (not self.params.same_offnoi_params(offnoi_dir)):
            self._logger.error('Parameters in offnoi directory do not match')
            return
        try:
            #offset_raw is quite big. deleted after use
            self.offset_raw = af.get_array_from_file(
                offnoi_dir, 'offset_raw.npy')
            self._logger.debug(self.offset_raw.shape)
            if self.offset_raw is None:
                self._logger.error('Error loading offset_raw data\n')
                return
            self.offset_fitted = af.get_array_from_file(
                offnoi_dir, 'offnoi_fit.npy'
            )[1]
            if self.offset_fitted is None:
                self._logger.error('Error loading fitted_offset data\n')
                return
            self.noise_fitted = af.get_array_from_file(
                offnoi_dir, 'offnoi_fit.npy'
            )[2]
            if self.noise_fitted is None:
                self._logger.error('Error loading fitted_noise data\n')
                return
            self.offnoi_dir = offnoi_dir
            self.common_dir = os.path.dirname(offnoi_dir)
            timestamp = datetime.now().strftime('%Y%m%d-%H%M%S')
            self.step_dir = os.path.join(
                self.common_dir,f'filter_{self.thres_event}_threshold'
            )
            self._logger.debug(self.step_dir)
        except:
            raise ValueError('Error loading offnoi data\n')

    def calculate(self):
        #create the working directory for the filter step
        os.makedirs(self.step_dir, exist_ok=True)
        self._logger.info(f'Created directory for filter step: {self.step_dir}')
        # and save the parameter file there
        self.params.save(os.path.join(self.step_dir, 'parameters.json'))

        data = an.get_data(self.bin_file, self.column_size, self.row_size, self.key_ints, self.nreps, self.nframes)
        gc.collect()
        if self.nreps_eval:
            data = an.exclude_nreps_eval(data, self.nreps_eval)
            self._logger.debug(f'Shape of data: {data.shape}')
        #omit bad pixels and mips frames
        if self.bad_pixels:
            data = an.set_bad_pixellist_to_nan(data, self.bad_pixels)
        if self.thres_bad_frames != 0:
            data = an.exclude_bad_frames(data, self.thres_bad_frames, self.step_dir)
            self._logger.debug(f'Shape of data: {data.shape}')
        if self.thres_mips != 0:
            data = an.exclude_mips_frames(data, self.thres_mips)
            self._logger.debug(f'Shape of data: {data.shape}')
        #offset the data and correct for common mode if necessary
        data = data - self.offset_raw[np.newaxis,:,:,:]
        self.offset_raw = None
        gc.collect()
        if self.comm_mode:
            an.correct_common_mode(data)
        if self.use_fitted_offset:
            #take care here, offset fitted can contain np.nan
            data -= np.nan_to_num(self.offset_fitted[np.newaxis,:,np.newaxis,:])
        avg_over_nreps = af.get_avg_over_nreps(data)
        np.save(os.path.join(self.step_dir, 'rndr_signals.npy'),
                avg_over_nreps)
        #calculate event map and save it
        event_map = an.calc_event_map(avg_over_nreps, self.noise_fitted, self.thres_event)
        np.save(os.path.join(self.step_dir, 'event_map.npy'),
                event_map)
        np.save(os.path.join(self.step_dir, 'sum_of_event_signals.npy'),
                an.get_sum_of_event_signals(event_map, self.row_size, self.column_size))
        np.save(os.path.join(self.step_dir, 'sum_of_event_counts.npy'),
                an.get_sum_of_event_counts(event_map, self.row_size, self.column_size))
        if self.thres_bad_slopes != 0:
            bad_slopes_pos, bad_slopes_data, bad_slopes_value = an.get_bad_slopes(data, self.thres_bad_slopes, self.step_dir)
            np.save(os.path.join(self.step_dir, 'bad_slopes_pos.npy'), bad_slopes_pos)
            np.save(os.path.join(self.step_dir, 'bad_slopes_data.npy'), bad_slopes_data)
            np.save(os.path.join(self.step_dir, 'bad_slopes_value.npy'), bad_slopes_value)

class Gain():

    _logger = logger.Logger('nproan-gain', 'debug').get_logger()

    def __init__(self, prm_file, filter_dir):
        self.load(prm_file, filter_dir)
        self._logger.info('Gain object created')

    def load(self, prm_file, filter_dir):
        self.params = pm.Params(prm_file)
        self.params_dict = self.params.get_dict()
        #common parameters
        self.results_dir = self.params_dict['common_results_dir']
        self.column_size = self.params_dict['common_column_size']
        self.row_size = self.params_dict['common_row_size']
        self.key_ints = self.params_dict['common_key_ints']
        self.bad_pixels = self.params_dict['common_bad_pixels']

        #gain parameters
        self.nreps = self.params_dict['filter_nreps']
        self.nframes = self.params_dict['filter_nframes']
        self.min_signals = self.params_dict['gain_min_signals']

        self._logger.info(f'Parameters loaded:')
        self.params.print_contents()
        
        self._logger.info('Checking parameters in filter directory')
        #look for a json file in the filter directory
        if (not self.params.same_common_params(filter_dir)) \
            or (not self.params.same_offnoi_params(filter_dir) \
            or (not self.params.same_filter_params(filter_dir))):
            self._logger.error('Parameters in filter directory do not match')
            return
        try:
            self.event_map = af.get_array_from_file(
                filter_dir, 'event_map.npy')
            #set the directory where the filter data is stored
            self.filter_dir = filter_dir
            #this is the parent directory. data from this step is stored there
            self.common_dir = os.path.dirname(filter_dir)
            timestamp = datetime.now().strftime('%Y%m%d-%H%M%S')
            self.step_dir = os.path.join(
                self.common_dir, timestamp + f'_gain_{self.min_signals}_minsignals'
            )
        except:
            self._logger.error('Error loading filter data\n')
            return
        self._logger.info('Filter data loaded\n')

    def calculate(self):
        #create the working directory for the gain step
        self.step_dir = os.path.join(self.common_dir, 
                                     f'gain_{self.min_signals}_min_signals')
        os.makedirs(self.step_dir, exist_ok=True)
        self._logger.info(f'Created directory for gain step: {self.step_dir}')
        # and save the parameter file there
        self.params.save(os.path.join(self.step_dir, 'parameters.json'))
        
        fits = an.get_gain_fit(self.event_map, self.row_size, self.column_size, self.min_signals)
        np.save(os.path.join(self.step_dir, 'fit_mean.npy'), fits[0])
        np.save(os.path.join(self.step_dir, 'fit_sigma.npy'), fits[1])
        np.save(os.path.join(self.step_dir, 'fit_mean_error.npy'), fits[2])
        np.save(os.path.join(self.step_dir, 'fit_sigma_error.npy'), fits[3])


import gc

from scipy.optimize import curve_fit
import numpy as np
from iminuit import cost, Minuit

from . import logger

_logger = logger.Logger(__name__, 'info').get_logger()

def fit_gauss_to_hist(data_to_fit: np.ndarray) -> np.ndarray:
    ''' 
    fits a gaussian to a histogram of data_to_fit
    Args:
        1D np.array
    Returns:
        np.array[amplitude, mean, sigma, error_amplitude, error_mean, error_sigma]
    '''
    guess = [1, np.nanmedian(data_to_fit), np.nanstd(data_to_fit)]
    try:
        hist, bins = np.histogram(data_to_fit, bins=100, range=(np.nanmin(data_to_fit), np.nanmax(data_to_fit)), density=True)
        bin_centers = (bins[:-1] + bins[1:]) / 2
        params, covar = curve_fit(gaussian, bin_centers, hist, p0=guess)
        return np.array([params[0],
                         params[1], 
                         np.abs(params[2]),
                         np.sqrt(np.diag(covar))[0],
                         np.sqrt(np.diag(covar))[1], 
                         np.sqrt(np.diag(covar))[2]])
    except:
        _logger.debug('Fitting for this histogram failed. Returning NaNs.')
        return np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan])
    
def unbinned_fit_gauss_to_hist(data_to_fit: np.ndarray) -> np.ndarray:
    '''
    fits a gaussian to a histogram of data_to_fit
    Args:
        1D np.array
    Returns:
        np.array[amplitude, mean, sigma, error_mean, error_sigma]
    '''
    #TODO: this doesnt seem to work: test this!
    c = cost.UnbinnedNLL(data_to_fit, gaussian)
    m = Minuit(c, 
               a1=1, 
               mu1=np.nanmedian(data_to_fit), 
               sigma1=np.nanstd(data_to_fit))
    m.limits['a1'] = (0, 100)
    m.limits['mu1'] = (np.nanmin(data_to_fit), np.nanmax(data_to_fit))
    m.limits['sigma1'] = (0, 100)
    m.migrad()
    if not m.valid:
        return np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan])
    return np.array([m.values['a1'],
                     m.values['mu1'],
                     np.abs(m.values['sigma1']),
                     m.errors['a1'],
                     m.errors['mu1'],
                     m.errors['sigma1']])

def get_fit_gauss(data: np.ndarray) -> np.ndarray:
    ''' 
    fits a gaussian to a histogram of data_to_fit
    using the scipy curve_fit method
    Args:
        np.array in shape (nframes, column_size, row_size)
    Returns:
        np.array in shape (6, rows, columns)
        index 0: amplitude
        index 1: mean
        index 2: sigma
        index 3: error_amplitude
        index 4: error_mean
        index 5: error_sigma
    '''
    if data.ndim != 3:
        _logger.error('Data is not a 3D array')
        raise ValueError('Data is not a 3D array')
    #apply the function to every frame
    output = np.apply_along_axis(fit_gauss_to_hist, axis = 0, arr = data)
    return output

def get_unbinned_fit_gauss(data: np.ndarray) -> np.ndarray:
    '''
    fits a gaussian to a histogram of data
    using the unbinned method in minuit
    returns a np.array in shape (6, rows, columns)
    Args:
        np.array in shape (nframes, column_size, row_size)
    Returns:
        np.array in shape (6, rows, columns)
        index 0: amplitude
        index 1: mean
        index 2: sigma
        index 3: error_amplitude
        index 4: error_mean
        index 5: error_sigma
    '''
    if data.ndim != 3:
        _logger.error('Data is not a 3D array')
        raise ValueError('Data is not a 3D array')
    #apply the function to every frame
    output = np.apply_along_axis(unbinned_fit_gauss_to_hist, axis = 0, arr = data)
    return output

def gaussian(x: float, a1: float, mu1: float, sigma1: float) -> float:
    return (a1 * np.exp(-(x - mu1)**2 / (2 * sigma1**2)))

def two_gaussians(x: float, a1: float, mu1: float, sigma1: float, 
                  a2: float, mu2: float, sigma2: float) -> float:
    return (a1 * np.exp(-(x - mu1)**2 / (2 * sigma1**2) +
            a2 * np.exp(-(x - mu2)**2 / (2 * sigma2**2))))

def linear_fit(data: np.ndarray) -> np.ndarray:
    x = np.arange(data.size)
    k, d = np.polyfit(x, data, 1)
    return np.array([k, d])


import numpy as np
from numba import njit, prange

from . import logger

_logger = logger.Logger(__name__, 'info').get_logger()

'''
Add parallized versions here.
Add nanmedian and nanmean for now.
Test if its good to always use parallel functions.
Write them only for keepdims=false and use np.newaxis if neccesary.
move to analysis funcs when they work.
'''

def nanmedian(data: np.ndarray, 
                       axis: int, 
                       keepdims: bool = False) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=axis, keepdims=keepdims).
    Args:
        data: np.array
        axis: int
        keepdims: bool
    Returns:
        np.array
    '''
    if data.ndim == 2:
        if axis == 0:
            if keepdims:
                return _nanmedian_2d_axis0(data)[np.newaxis,:]
            else:
                return _nanmedian_2d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmedian_2d_axis1(data)[:,np.newaxis]
            else:
                return _nanmedian_2d_axis1(data)
    elif data.ndim == 3:
        if axis == 0:
            if keepdims:
                return _nanmedian_3d_axis0(data)[np.newaxis,:,:]
            else:
                return _nanmedian_3d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmedian_3d_axis1(data)[:,np.newaxis,:]
            else:
                return _nanmedian_3d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmedian_3d_axis2(data)[:,:,np.newaxis]
            else:
                return _nanmedian_3d_axis2(data)
    elif data.ndim == 4:
        if axis == 0:
            if keepdims:
                return _nanmedian_4d_axis0(data)[np.newaxis,:,:,:]
            else:
                return _nanmedian_4d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmedian_4d_axis1(data)[:,np.newaxis,:,:]
            else:
                return _nanmedian_4d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmedian_4d_axis2(data)[:,:,np.newaxis,:]
            else:
                return _nanmedian_4d_axis2(data)
        elif axis == 3:
            if keepdims:
                return _nanmedian_4d_axis3(data)[:,:,:,np.newaxis]
            else:
                return _nanmedian_4d_axis3(data)
    else:
        _logger.error('Data has wrong dimensions')
        return

@njit(parallel=True)
def _nanmedian_4d_axis0(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=0, keepdims=False).
    Args:
        data: 4D np.array 
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_1_size,axis_2_size, axis_3_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmedian(data[:,i,j,k])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmedian_4d_axis1(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=1, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size,axis_2_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmedian(data[i,:,j,k])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmedian_4d_axis2(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=2, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size,axis_1_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_3_size):
                median = np.nanmedian(data[i,j,:,k])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmedian_4d_axis3(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=3, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]  
    output = np.zeros((axis_0_size,axis_1_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_2_size):
                median = np.nanmedian(data[i,j,k,:])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmedian_3d_axis0(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=0, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    '''
    if data.ndim != 3:
        _logger.error('Input data is not a 3D array.')
        raise ValueError('Input data is not a 3D array.')
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_1_size, axis_2_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            median = np.nanmedian(data[:,i,j])
            output[i,j] = median
    return output

@njit(parallel=True)
def _nanmedian_3d_axis1(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=1, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    '''
    if data.ndim != 3:
        _logger.error('Input data is not a 3D array.')
        raise ValueError('Input data is not a 3D array.')
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_0_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            median = np.nanmedian(data[i,:,j])
            output[i,j] = median
    return output

@njit(parallel=True)
def _nanmedian_3d_axis2(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=2, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    '''
    if data.ndim != 3:
        _logger.error('Input data is not a 3D array.')
        raise ValueError('Input data is not a 3D array.')
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    output = np.zeros((axis_0_size, axis_1_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            median = np.nanmedian(data[i,j,:])
            output[i,j] = median
    return output

@njit(parallel=True)
def _nanmedian_2d_axis0(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmedian(data, axis=0, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    '''
    if data.ndim != 2:
        _logger.error('Input data is not a 2D array.')
        raise ValueError('Input data is not a 2D array.')
    axis_1_size = data.shape[1]
    output = np.zeros(axis_1_size)
    for i in prange(axis_1_size):
        median = np.nanmedian(data[:,i])
        output[i] = median
    return output

@njit(parallel=True)
def _nanmedian_2d_axis1(data: np.ndarray) -> np.ndarray:

    '''
    The equivalent to np.nanmedian(data, axis=1, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    '''
    if data.ndim != 2:
        _logger.error('Input data is not a 2D array.')
        raise ValueError('Input data is not a 2D array.')
    axis_0_size = data.shape[0]
    output = np.zeros(axis_0_size)
    for i in prange(axis_0_size):
        median = np.nanmedian(data[i,:])
        output[i] = median
    return output

def nanmean(data: np.ndarray, 
                       axis: int, 
                       keepdims: bool = False) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=axis, keepdims=keepdims).
    Args:
        data: np.array
        axis: int
        keepdims: bool
    Returns:
        np.array
    '''
    if data.ndim == 2:
        if axis == 0:
            if keepdims:
                return _nanmean_2d_axis0(data)[np.newaxis,:]
            else:
                return _nanmean_2d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmean_2d_axis1(data)[:,np.newaxis]
            else:
                return _nanmean_2d_axis1(data)
    elif data.ndim == 3:
        if axis == 0:
            if keepdims:
                return _nanmean_3d_axis0(data)[np.newaxis,:,:]
            else:
                return _nanmean_3d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmean_3d_axis1(data)[:,np.newaxis,:]
            else:
                return _nanmean_3d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmean_3d_axis2(data)[:,:,np.newaxis]
            else:
                return _nanmean_3d_axis2(data)
    elif data.ndim == 4:
        if axis == 0:
            if keepdims:
                return _nanmean_4d_axis0(data)[np.newaxis,:,:,:]
            else:
                return _nanmean_4d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmean_4d_axis1(data)[:,np.newaxis,:,:]
            else:
                return _nanmean_4d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmean_4d_axis2(data)[:,:,np.newaxis,:]
            else:
                return _nanmean_4d_axis2(data)
        elif axis == 3:
            if keepdims:
                return _nanmean_4d_axis3(data)[:,:,:,np.newaxis]
            else:
                return _nanmean_4d_axis3(data)
    else:
        _logger.error('Data has wrong dimensions')
        return

@njit(parallel=True)
def _nanmean_4d_axis0(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=0, keepdims=False).
    Args:
        data: 4D np.array 
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_1_size,axis_2_size, axis_3_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmean(data[:,i,j,k])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmean_4d_axis1(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=1, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size,axis_2_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmean(data[i,:,j,k])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmean_4d_axis2(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=2, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size,axis_1_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_3_size):
                median = np.nanmean(data[i,j,:,k])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmean_4d_axis3(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=3, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    '''
    if data.ndim != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]  
    output = np.zeros((axis_0_size,axis_1_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_2_size):
                median = np.nanmean(data[i,j,k,:])
                output[i,j,k] = median
    return output

@njit(parallel=True)
def _nanmean_3d_axis0(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=0, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    '''
    if data.ndim != 3:
        _logger.error('Input data is not a 3D array.')
        raise ValueError('Input data is not a 3D array.')
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_1_size, axis_2_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            median = np.nanmean(data[:,i,j])
            output[i,j] = median
    return output

@njit(parallel=True)
def _nanmean_3d_axis1(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=1, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    '''
    if data.ndim != 3:
        _logger.error('Input data is not a 3D array.')
        raise ValueError('Input data is not a 3D array.')
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_0_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            median = np.nanmean(data[i,:,j])
            output[i,j] = median
    return output

@njit(parallel=True)
def _nanmean_3d_axis2(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=2, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    '''
    if data.ndim != 3:
        _logger.error('Input data is not a 3D array.')
        raise ValueError('Input data is not a 3D array.')
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    output = np.zeros((axis_0_size, axis_1_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            median = np.nanmean(data[i,j,:])
            output[i,j] = median
    return output

@njit(parallel=True)
def _nanmean_2d_axis0(data: np.ndarray) -> np.ndarray:
    '''
    The equivalent to np.nanmean(data, axis=0, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    '''
    if data.ndim != 2:
        _logger.error('Input data is not a 2D array.')
        raise ValueError('Input data is not a 2D array.')
    axis_1_size = data.shape[1]
    output = np.zeros(axis_1_size)
    for i in prange(axis_1_size):
        median = np.nanmean(data[:,i])
        output[i] = median
    return output

@njit(parallel=True)
def _nanmean_2d_axis1(data: np.ndarray) -> np.ndarray:
    
    '''
    The equivalent to np.nanmean(data, axis=1, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    '''
    if data.ndim != 2:
        _logger.error('Input data is not a 2D array.')
        raise ValueError('Input data is not a 2D array.')
    axis_0_size = data.shape[0]
    output = np.zeros(axis_0_size)
    for i in prange(axis_0_size):
        median = np.nanmean(data[i,:])
        output[i] = median
    return output


import os
import fnmatch
import json

from . import logger

class Params:

    _logger = logger.Logger(__name__, 'info').get_logger()

    '''
    To change/add parameters, edit/add them here.
    Also change the load() function in the respective class.
    '''
    common_params = {
        'common_results_dir': "",           #str
        'common_column_size': 64,           #int
        'common_row_size': 64,              #int
        'common_key_ints': 3,               #int
        'common_bad_pixels': []             #list of tuples
    }
    offnoi_params = {
        'offnoi_bin_file': "",              #str
        'offnoi_nreps': 200,                 #int
        'offnoi_nframes': 100,                #int
        'offnoi_nreps_eval': [0,-1,1],        #list of ints
        'offnoi_comm_mode': True,             #bool
        'offnoi_thres_mips': 1000,            #float
        'offnoi_thres_bad_frames': 5,          #float
        'offnoi_thres_bad_slopes': 5          #float
    }
    filter_params = {
        'filter_bin_file': "",            #str
        'filter_nreps': 200,               #int
        'filter_nframes': 100,              #int
        'filter_nreps_eval': [0,-1,1],           #list of ints
        'filter_comm_mode': True,           #bool
        'filter_thres_mips': 1000,          #float
        'filter_thres_event': 5,            #float
        'filter_use_fitted_offset': False,  #bool
        'filter_thres_bad_frames': 5,        #float
        'filter_thres_bad_slopes': 5        #float
    }
    gain_params = {
        'gain_min_signals': 5               #int
    }

    #types are checked when they are read
    params_types = {
        'common_results_dir': str,
        'common_column_size': int,
        'common_row_size': int,
        'common_key_ints': int,
        'common_bad_pixels': list,          #actually, list of tuples

        'offnoi_bin_file': str,
        'offnoi_nreps': int,
        'offnoi_nframes': int,
        'offnoi_nreps_eval': list,
        'offnoi_comm_mode': bool,
        'offnoi_thres_mips': (int, float),
        'offnoi_thres_bad_frames': (int, float),
        'offnoi_thres_bad_slopes': (int, float),

        'filter_bin_file': str,
        'filter_nreps': int,
        'filter_nframes': int,
        'filter_nreps_eval': list,
        'filter_comm_mode': bool,
        'filter_thres_mips': (int, float),
        'filter_thres_event': (int, float),
        'filter_use_fitted_offset': bool,
        'filter_thres_bad_frames': (int, float),
        'filter_thres_bad_slopes': (int, float),

        'gain_min_signals': int
    }

    #required parameters, where there is no default value
    #file cannot be loaded if these are missing
    required_params = [
        'offnoi_bin_file',
        'offnoi_nreps',
        'filter_bin_file',
        'filter_nreps'
    ]

    def __init__(self, json_path: str = None):
        self.default_dict = {**self.common_params,
                             **self.offnoi_params,
                             **self.filter_params,
                             **self.gain_params}
        self.inp_dict = None
        self.param_dict = None
        if json_path is not None:
            self.update(json_path)
            self.check_types()
        else:
            self._logger.error('No parameter file provided.')
            self._logger.error('Run save_default_file() to save a default parameter file.')
            self._logger.error('Add all required parameters to default and start new.')

    def update(self, json_path: str) -> None:
        try:
            with open(json_path) as f:
                self.inp_dict = json.load(f)
        except:
            self._logger.error('Error loading the parameter file.')
            self.save_default_file()
            self._logger.error('A default parameter file has been saved to the current directory.')
            self.param_dict = None
            return
        self.param_dict = self.default_dict.copy()
        #check consistency of the input dict with the default dict
        for key,value in self.inp_dict.items():
            if key not in self.default_dict:
                self._logger.error(f"{key} is not a valid parameter.")
            else:
                self.param_dict[key] = value
        #check for missing parameters, using default if not required
        #if parameter has no default, set param_dict to None
        for key,value in self.param_dict.items():
            if value is None:
                if key in self.required_params:
                    self._logger.error(f"{key} is missing in the file.")
                    self._logger.error('Please provide a complete parameter file')
                    self._logger.error('Run .info() to see the required parameters.')
                    self.param_dict = None
                    break
                else:
                    self._logger.error(f"{key} is missing. Using default: {self.default_dict[key]}")

    def check_types(self) -> None:
        for key,value in self.param_dict.items():
            if key not in self.params_types:
                raise TypeError(f"There is no type defined for {key}.")
            else:
                expected_type = self.params_types[key]
                if not isinstance(value, expected_type):
                    raise TypeError(f"Expected {key} to be of type {expected_type}.")
                
    def get_dict(self) -> dict:
        return self.param_dict

    def print_contents(self) -> None:
        for key, value in self.param_dict.items():
            self._logger.info(f"{key}: {value}")

    def info(self) -> None:
        self._logger.info('The following parameters must be provided:')
        self._logger.info('--common parameters:')
        for key in self.common_params.keys():
            self._logger.info(key)
        self._logger.info('--offnoi parameters:')
        for key in self.offnoi_params.keys():
            self._logger.info(key)
        self._logger.info('--filter parameters:')
        for key in self.filter_params.keys():
            self._logger.info(key)
        self._logger.info('--gain parameters:')
        for key in self.gain_params.keys():
            self._logger.info(key)
        self._logger.info('required parameters:')
        for key in self.required_params:
            self._logger.info(key)

    def save_default_file(self, path: str = None) -> None:
        #if no path is provided, save to the current directory
        self._logger.info(f'path: {path}')
        if path is None:
            path = os.path.join(os.getcwd(), 'default_params.json')
        else:
            path = os.path.join(path, 'default_params.json')
        with open(path, 'w') as f:
            json.dump(self.default_dict, f, indent=4)
    
    def save(self, path: str) -> None:
        with open(path, 'w') as f:
            json.dump(self.param_dict, f, indent=4)

    def get_json_file_name_in_folder(self, folder_path: str) -> str:
        count = 0
        json_file =''
        for file in os.listdir(folder_path):
            if fnmatch.fnmatch(file, '*.json'):
                json_file = os.path.join(folder_path, file)
                count += 1
        if count == 1:
            return json_file
        else:
            return None
            
    def same_common_params(self, folder_path: str) -> bool:
        json_file = self.get_json_file_name_in_folder(folder_path)
        self._logger.info(f'json_file: {json_file}')
        if json_file:
            with open(json_file) as f:
                dict = json.load(f)
            for key in self.common_params.keys():
                if self.param_dict[key] != dict[key]:
                    return False
            return True
        else:
            self._logger.error('No json file found in the folder')
            return False
        
    def same_offnoi_params(self, folder_path:str) -> bool:
        json_file = self.get_json_file_name_in_folder(folder_path)
        if json_file:
            with open(json_file) as f:
                dict = json.load(f)
            for key in self.offnoi_params.keys():
                if self.param_dict[key] != dict[key]:
                    return False
            return True
        else:
            self._logger.error('No json file found in the folder')
            return False
    
    def same_filter_params(self, folder_path:str) -> bool:
        json_file = self.get_json_file_name_in_folder(folder_path)
        if json_file:
            with open(json_file) as f:
                dict = json.load(f)
            for key in self.filter_params.keys():
                if self.param_dict[key] != dict[key]:
                    return False
            return True
        else:
            self._logger.error('No json file found in the folder')
            return False
        
    def same_gain_params(self, folder_path: str) -> bool:
        json_file = self.get_json_file_name_in_folder(folder_path)
        if json_file:
            with open(json_file) as f:
                dict = json.load(f)
            for key in self.gain_params.keys():
                if self.param_dict[key] != dict[key]:
                    return False
            return True
        else:
            self._logger.error('No json file found in the folder')
            return False
        
    def estimate_ram_usage(self) -> None:
        pixel_per_frame = self.param_dict['common_column_size'] * self.param_dict['common_row_size']
        frames_offnoi = self.param_dict['offnoi_nframes']
        frames_filter = self.param_dict['filter_nframes']
        nreps_offnoi = self.param_dict['offnoi_nreps']
        nreps_filter = self.param_dict['filter_nreps']
        offnoi = int((frames_offnoi * nreps_offnoi * pixel_per_frame * 8) /(1024**3))
        filter = int((frames_filter * nreps_filter * pixel_per_frame * 8) /(1024**3))

        self._logger.info(f'Recommended RAM for OffNoi: {offnoi*3+2} GB')
        self._logger.info(f'Recommended RAM for Filter: {filter*3+2} GB')


import gc

import numpy as np
from scipy.optimize import curve_fit

from . import logger
from . import fitting
from . import display
from . import parallel_funcs

_logger = logger.Logger(__name__, 'info').get_logger()

def get_data(bin_file : str, column_size: int, row_size: int, key_ints: int,
             nreps: int, nframes: int) -> np.ndarray:
        '''
        Reads the binary file (in chunks) and returns the data as a numpy array 
        in shape (nframes, column_size, nreps, row_size).
        Args:
            bin_file: path to the binary file
            column_size: number of columns in the data
            row_size: number of rows in the data
            key_ints: number of key ints in the data
            nreps: number of repetitions in the data
            nframes: number of frames to read from the data    
        Returns:
            np.array in shape (nframes, column_size, nreps, row_size)
        '''
        polarity = -1
        frames_per_chunk = 20
        raw_row_size = column_size + key_ints
        raw_frame_size = column_size * raw_row_size * nreps
        chunk_size =  raw_frame_size* frames_per_chunk
        offset = 8
        count = 0
        output = np.zeros((nframes, column_size, nreps, 
                              row_size), dtype=np.float64)
        frames_here = 0

        while True:
            inp_data = np.fromfile(bin_file, dtype='uint16', 
                                   count = chunk_size, offset = offset)
            #check if file is at its end
            if inp_data.size == 0:
                _logger.warning(f'Loaded {frames_here} of {nframes} requested frames, end of file reached.')
                return polarity*output[:frames_here].copy()
            _logger.info(f'Reading chunk {count+1} from bin file, {frames_here} frames loaded')
            #reshape the array into rows -> (#ofRows,67)
            inp_data = inp_data.reshape(-1, raw_row_size)
            #find all the framekeys
            frame_keys = np.where(inp_data[:, column_size] == 65535)
            #stack them and calculate difference to find incomplete frames
            frames = np.stack((frame_keys[0][:-1], frame_keys[0][1:]))
            del frame_keys
            gc.collect()
            diff = np.diff(frames, axis = 0)
            rows_per_frame = row_size*nreps
            valid_frames_position = np.nonzero(diff == rows_per_frame)[1]
            if len(valid_frames_position) == 0:
                _logger.error('No valid frames found in chunk, wrong nreps.')
                raise Exception('No valid frames found in chunk, wrong nreps.')
            del diff
            gc.collect()
            valid_frames = frames.T[valid_frames_position]
            frame_start_indices = valid_frames[:,0]
            frame_end_indices = valid_frames[:,1]
            del valid_frames
            gc.collect()
            offset += chunk_size * 2
            count += 1
            inp_data = np.array([inp_data[start+1:end+1, :64] 
                                 for start, end 
                                 in zip(frame_start_indices, 
                                        frame_end_indices)])
            frames_inc = inp_data.shape[0]
            #check if inp_data would make data full
            if frames_here + frames_inc > nframes:
                inp_data = inp_data.reshape(-1, column_size, 
                                             nreps, 
                                             row_size).astype(float)
                output[frames_here:] = inp_data[:nframes-frames_here]
                frames_here += frames_inc
                _logger.info(f'\nLoaded {nframes} frames')
                return polarity*output.copy()
            output[frames_here:frames_here+frames_inc] = \
            inp_data.reshape(-1, column_size, 
                              nreps, row_size).astype(float)
            frames_here += frames_inc
            gc.collect()

def get_dummy_data(column_size: int, row_size: int, nreps: int, 
                   nframes: int) -> np.ndarray:
    '''
    Returns random values for tests.
    Args:
        column_size: number of columns in the data
        row_size: number of rows in the data
        key_ints: number of key ints in the data
        nreps: number of repetitions in the data
        nframes: number of frames to read from the data
        
    Returns:
        np.array in shape (nframes, column_size, nreps, row_size)
    '''
    return np.random.rand(nframes, row_size, nreps, column_size)

def exclude_nreps_eval(data: np.ndarray, nreps_eval: list) -> np.ndarray:
    '''
    Deletes nreps from data that are not in the list nreps_eval.
    nreps_eval is a list of 3 integers: [lower, upper, step]
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
        nreps_eval: list of 3 ints
    Returns:
        np.array View(!) (nframes, column_size, nreps-X, row_size)
    '''
    #TODO: optimize this
    if np.ndim(data) != 4:
        _logger.error('Data has wrong dimensions')
        return None
    if len(nreps_eval) != 3:
        _logger.error('nreps_eval must be a list of 3 integers')
        raise ValueError('nreps_eval must be a list of 3 integers')
    lower = nreps_eval[0]
    upper = nreps_eval[1]
    step = nreps_eval[2]
    if upper == -1:
        upper = data.shape[2]
    if lower < 0:
        raise ValueError('Lower limit must be greater or equal 0')
    if upper > data.shape[2]:
        raise ValueError('Upper limit is greater than the number of nreps')
    if upper < lower:
        raise ValueError('Upper limit must be greater than lower limit')
    _logger.info('Excluding nreps')
    mask = np.zeros(data.shape[2])
    mask[lower:upper:step] = 1
    mask = mask.astype(bool)
    _logger.info(f'Excluded {np.sum(~mask)} nreps')
    return data[:,:,mask,:]

def exclude_mips_frames(data: np.ndarray, thres_mips: int) -> np.ndarray:
    '''
    Calculates the median of each frame and deletes frames that are
    above or below the median by a certain threshold.
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
        thres_mips: absolute threshold in adu
    Returns:
        np.array (nframes-X, column_size, nreps, row_size)
    '''
    #TODO add optional inputs and use parallel stuff for median
    #TODO: combine with exclude_bad_frames to save computation of median
    if np.ndim(data) != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    _logger.info(f'Excluding bad frames due to MIPS, threshold: {thres_mips}')
    median = np.nanmedian(data, 
                          axis = (1,2,3))[:,np.newaxis,np.newaxis,np.newaxis]
    mask = (data > median + thres_mips) | (data < median - thres_mips)
    del median
    gc.collect()
    mask = np.any(mask, axis = (1,2,3))
    _logger.info(f'Excluded {np.sum(mask)} frames')
    return data[~mask]

def exclude_bad_frames(data: np.ndarray, thres_bad_frames: int, 
                       step_dir: str = None):
    '''
    Calculates the average of each frame and excludes frames that are
    above or below the fitted mean by a certain threshold.
    It saves a .png file in the step directory.
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
        thres_bad_frames: used with the fitted sigma do exclude frames
        step_dir (optional): directory in which plot is saved
    Returns:
        np.array in shape (nframes-X, column_size, nreps, row_size)
    '''
    #TODO: optimize this
    if np.ndim(data) != 4:
        _logger.error('Data has wrong dimensions')
        return None
    _logger.info('Excluding bad frames')
    avg_per_frame = np.nanmean(data, axis = (1,2,3))
    fit = fitting.fit_gauss_to_hist(avg_per_frame)
    lower_bound = fit[1] - thres_bad_frames*np.abs(fit[2])
    upper_bound = fit[1] + thres_bad_frames*np.abs(fit[2])
    bad_pixel_mask = (avg_per_frame < lower_bound) | (avg_per_frame > upper_bound)
    excluded_frames = np.sum(bad_pixel_mask)
    _logger.info(f'Excluded {excluded_frames} frames')
    title = f'Average signal per frame. Excluded {excluded_frames} frames'
    display.draw_hist_and_gauss_fit(avg_per_frame, 100, fit[0], fit[1], fit[2],
                            'bad_frames', 
                            save_to = step_dir)
    return data[~bad_pixel_mask]

def get_bad_slopes(data: np.ndarray, thres_bad_slopes: int, 
                   step_dir: str = None)-> tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculates the slope over nreps for every pixel and frame.
    It then fits a gaussian to the histogram of the slopes, and determines
    the bad slopes by a threshold.
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
        thres_bad_slopes: used with the fitted sigma to determine bad slopes
        step_dir (optional): directory in which plot is saved
    Returns:
        np.array in shape (n, 3) with the position [frame, row, column]
        np.array in shape (n, nreps) with the data of the bad slopes
    '''
    #TODO: optimize this by doing the linear fit in parallel
    if np.ndim(data) != 4:
        _logger.error('Input data is not a 4D array.')
        raise ValueError('Input data is not a 4D array.')
    _logger.info('Calculating bad slopes')
    slopes = np.apply_along_axis(fitting.linear_fit, axis = 2, arr = data)[:, :, 0, :]
    _logger.debug(f'Shape of slopes: {slopes.shape}')
    fit = fitting.fit_gauss_to_hist(slopes.flatten())
    _logger.debug(f'Fit: {fit}')
    lower_bound = fit[1] - thres_bad_slopes*np.abs(fit[2])
    upper_bound = fit[1] + thres_bad_slopes*np.abs(fit[2])
    _logger.debug(f'Lower bound: {lower_bound}, Upper bound: {upper_bound}')
    bad_slopes_mask = (slopes < lower_bound) | (slopes > upper_bound)
    frame, row, column = np.where(bad_slopes_mask)
    bad_slopes_pos = np.array([frame, row, column]).T
    bad_slopes_value = slopes[bad_slopes_mask]
    #get indices of frames with bad slopes
    bad_slopes_data = data[frame.T, row.T, :, column.T]
    _logger.info(f'Found {len(bad_slopes_pos)} bad Slopes')
    _logger.debug(f'Shape of bad slopes data: {bad_slopes_data.shape}')
    _logger.debug(f'Shape of bad slopes pos: {bad_slopes_pos.shape}')
    title = f'Slope Values for each Pixel and Frame. {len(bad_slopes_pos)} bad slopes found.'
    display.draw_hist_and_gauss_fit(slopes.flatten(), 100, fit[0], fit[1], fit[2],
                            'bad_slopes', 
                            save_to = step_dir)
    return bad_slopes_pos, bad_slopes_data, bad_slopes_value

def set_bad_pixellist_to_nan(data: np.ndarray, 
                             bad_pixels: list[tuple[int, int]]) -> None:
    '''
    Sets all ignored Pixels in data to NaN. List of pixels is from the
    parameter file. [(row,col), (row,col), ...]
    Args:
        data: np.array (nframes, column_size, nreps, row_size)
        bad_pixels: list of tuples (row,col)
    Returns:
        np.array (nframes, column_size, nreps, row_size)
    '''
    if np.ndim(data) != 4:
        _logger.error('Data is not a 4D array')
        raise ValueError('Data is not a 4D array')
    _logger.info('Excluding bad pixels')
    bad_pixel_mask = np.zeros(data.shape, dtype=bool)
    for index in bad_pixels:
        col = index[1]
        row = index[0]
        bad_pixel_mask[:,row,:,col] = True
    data[bad_pixel_mask] = np.nan
    _logger.info(f'Excluded {len(bad_pixels)} pixels')

def correct_common_mode(data: np.ndarray) -> None:
    '''
    Calculates the median of euch row in data, and substracts it from 
    the row.
    Correction is done inline to save memory.
    Args:
        np.array in shape (nframes, column_size, nreps, row_size)
    '''
    if data.ndim != 4:
        _logger.error('Data is not a 4D array')
        raise ValueError('Data is not a 4D array')
    _logger.info('Starting common mode correction.')  
    # Iterate over the nframes dimension
    # Calculate the median for one frame
    #median_common = analysis_funcs.parallel_nanmedian_4d_axis3(data)
    # Subtract the median from the frame in-place
    median_common = parallel_funcs.nanmedian(data, axis=3, keepdims=True)
    data -= median_common
    _logger.info('Data is corrected for common mode.')

def calc_event_map(avg_over_nreps: np.ndarray, noise_fitted: np.ndarray, 
                   thres_event: int) -> np.ndarray:
    '''
    Finds events in the data by comparing the average over nreps with the
    noise fitted map.
    Args:
        avg_over_nreps: np.array (nframes, column_size, row_size)
        noise_fitted: np.array (column_size, row_size)
        thres_event: threshold for the events
    Returns:
        np.array (column_size, row_size) with lists of signals
    '''
    if avg_over_nreps.ndim != 3:
        _logger.error('avg_over_nreps is not a 3D array')
        raise ValueError('avg_over_nreps is not a 3D array')
    if noise_fitted.ndim != 2:
        _logger.error('noise_fitted is not a 2D array')
        raise ValueError('noise_fitted is not a 2D array')
    _logger.info('Finding events')
    threshold_map = noise_fitted * thres_event
    events = avg_over_nreps > threshold_map[np.newaxis,:,:]
    signals = avg_over_nreps[events]
    indices = np.transpose(np.where(events))
    _logger.info(f'{signals.shape[0]} events found')
    event_array = np.concatenate(
        (indices, signals[:,np.newaxis]),
          axis = 1
        )
    event_map = np.zeros((64,64), dtype = object)
    event_map.fill([])
    for entry in event_array:
        row = int(entry[1])
        column = int(entry[2])
        signal = entry[3]
        event_map[row][column] = np.append(
            event_map[row][column], signal
            )
    return event_map

def get_sum_of_event_signals(event_map: np.ndarray, 
                             row_size: int, 
                             column_size: int) -> np.ndarray:
    '''
    Adds the signal values of the events in the event_map
    Args:
        event_map: np.array (column_size, row_size) 
        row_size: number of rows in the data
        column_size: number of columns in the data
    Returns:
        np.array (column_size, row_size) with the sum of the signals
    '''
    sum_of_events = np.zeros((row_size,column_size))
    for row in range(row_size):
        for column in range(column_size):
            sum_of_events[row][column] = sum(event_map[row][column])
    return sum_of_events

def get_sum_of_event_counts(event_map: np.ndarray, 
                            row_size: int, column_size: int) -> np.ndarray:
    '''
    Sums up the number of events in the event_map
    Args:
        event_map: np.array (column_size, row_size) 
        row_size: number of rows in the data
        column_size: number of columns in the data
    Returns:
        np.array (column_size, row_size) with the count of the events
    '''
    sum_of_events = np.zeros((row_size,column_size))
    for row in range(row_size):
        for column in range(column_size):
            sum_of_events[row][column] = len(event_map[row][column])
    return sum_of_events

def get_gain_fit(event_map: np.ndarray, row_size: int, column_size: int, 
                 min_signals: int) -> tuple[np.ndarray, np.ndarray,np.ndarray,np.ndarray]:
    '''
    Fits a gaussian to the histogram of the signals in the event_map. When
    there are fewer signals than min_signals, or the fit is not
    successfull, the values are set to np.nan.
    Args:
        event_map: np.array (column_size, row_size) with lists of signals
        row_size: number of rows in the data
        column_size: number of columns in the data
        min_signals: minimum number of signals for a fit
    Returns:
        np.array (column_size, row_size) with the mean of the gaussian
        np.array (column_size, row_size) with the sigma of the gaussian
        np.array (column_size, row_size) with the error of the mean
        np.array (column_size, row_size) with the error of the sigma
        '''
    #TODO: optimize this
    mean = np.full((row_size, column_size), np.nan)
    sigma = np.full((row_size, column_size), np.nan)
    mean_error = np.full((row_size, column_size), np.nan)
    sigma_error = np.full((row_size, column_size), np.nan)

    def fit_hist(data):

        def gaussian(x, a1, mu1, sigma1):
            return a1 * np.exp(-(x - mu1)**2 / (2 * sigma1**2))
        try:
            hist, bins = np.histogram(data, bins=10, 
                                      range=(np.nanmin(data), np.nanmax(data)), 
                                      density=True)
            bin_centers = (bins[:-1] + bins[1:]) / 2
            params, covar = curve_fit(gaussian, 
                                      bin_centers, 
                                      hist, 
                                      p0=[1,bins[np.argmax(hist)],1])
            return (params[1], abs(params[2]), 
                    np.sqrt(np.diag(covar))[1], 
                    np.sqrt(np.diag(covar))[2])
        except:
            return (np.nan,np.nan,np.nan,np.nan)
    
    count_too_few = 0
    for i in range(event_map.shape[0]):
        for j in range(event_map.shape[1]):
            signals = event_map[i,j]
            if len(signals) >= min_signals:
                params = fit_hist(signals)
                mean[i,j] = params[0]
                sigma[i,j] = params[1]
                mean_error[i,j] = params[2]
                sigma_error[i,j] = params[3]
            else:
                count_too_few += 1
    _logger.info(f'{count_too_few} pixels have less than {min_signals} signals')
    return mean, sigma, mean_error, sigma_error


import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

from . import logger
from . import fitting

_logger = logger.Logger(__name__, 'info').get_logger()

def draw_hist(data: np.ndarray, file_name: str = "histogram", 
              save_to: str = None, **kwargs) -> None:
    '''
    Draw a histogram of the data.
    Args:
        np.array in any shape
        file_name: str
        save_to: str
        **kwargs: passed to plt.hist
    '''
    plt.clf()
    plt.hist(data.ravel(), **kwargs)
    if save_to is not None:
        plt.savefig(save_to + file_name + '.png')
        plt.close()
    else:
        plt.show()

def draw_heatmap(data: np.ndarray, file_name: str = "heatmap", 
                 save_to: str = None, **kwargs) -> None:
    '''
    Draw a heatmap of the data
    Args:
        np.array
        file_name: str
        save_to: str (optional)
        **kwargs: passed to sns.heatmap
    '''
    if data.ndim !=2:
        _logger.error('Input data is not a 2D array.')
        raise ValueError('Input data is not a 2D array.')
    plt.clf()
    cmap = kwargs.get('cmap', 'coolwarm')
    sns.heatmap(data, cmap=cmap, **kwargs)
    if save_to is not None:
        plt.savefig(save_to + file_name + '.png')
        plt.close()
    else:
        plt.show()

def draw_graph(data: np.ndarray, file_name: str = "graph", 
               save_to: str = None, **kwargs) -> None:
    '''
    Draw a graph of the data
    Args:
        np.array in any shape
        file_name: str
        save_to: str (optional)
        **kwargs: passed to plt.plot
    '''
    plt.clf()
    plt.plot(data.ravel(), **kwargs)
    if save_to is not None:
        plt.savefig(save_to + file_name + '.png')
        plt.close()
    else:
        plt.show()

def draw_hist_and_gauss_fit(data: np.ndarray, bins: str, 
                            amplitude: float, mean: float, sigma: float, 
                            file_name: str = None, 
                            save_to: str = None) -> None:
    '''
    Draw a histogram of the data and a gaussian fit
    Args:
        np.array
        bins: int
        amplitude: float
        mean: float
        sigma: float
        title: str
        file_name: str
        save_to: str (optional)
    '''
    plt.clf()
    hist, hist_bins = np.histogram(data, bins, range=(np.nanmin(data), np.nanmax(data)), density=True)
    bin_centers = (hist_bins[:-1] + hist_bins[1:]) / 2
    plt.hist(data, bins=hist_bins, density=True, alpha=0.5)
    plt.plot(bin_centers, fitting.gaussian(bin_centers, amplitude, mean, sigma), 'r-')
    plt.title(f'Fitted parameters:\nMean: {mean:.2f}\nSigma: {sigma:.2f}')
    plt.xlabel('Value')
    plt.ylabel('Frequency')
    if save_to is not None:
        if file_name is None:
            file_name = 'hist_and_fit'
        plt.savefig(save_to + '/' + file_name + '.png')
        plt.close()
    else:
        plt.show()